use std::collections::BTreeMap;
use std::num::NonZeroU32;

use crate::MeanSD;
use crate::StandardDeviationMode;

/// A bin as used in [`Binned`].
#[derive(Clone, Copy, Debug, Default, Eq, Ord, PartialEq, PartialOrd)]
pub struct Bin {
    bottom: i64,
    top: i64,
}

impl Bin {
    /// Returns the bottom end.
    #[must_use]
    pub const fn bottom(&self) -> i64 {
        self.bottom
    }

    /// Returns the top end.
    #[must_use]
    pub const fn top(&self) -> i64 {
        self.top
    }
}

/// An online algorithm for calculating mean and standard deviation with
/// binning and fixed bin sizes.
///
/// The memory requirement is effectively constant, growing only ever so
/// slightly with each new bin.
#[derive(Clone, Debug)]
pub struct Binned {
    all: MeanSD,
    data: BTreeMap<i64, MeanSD>,
    width: NonZeroU32,
}

/// Constructors.
impl Binned {
    /// Returns an empty `Binned` structure with `width` bin width.
    ///
    /// **Note:** The bins always use sample standard deviation, `sd_mode` only
    /// applies to the [`MeanSD`] for all observations.
    #[must_use]
    pub fn new(width: NonZeroU32, sd_mode: StandardDeviationMode) -> Self {
        Self {
            all: MeanSD::new(sd_mode),
            data: BTreeMap::new(),
            width,
        }
    }
}

/// Add new observations.
impl Binned {
    /// Updates the data structure with a new observation `x`.
    pub fn update(&mut self, x: f64) {
        self.all.update(x);

        let width = f64::from(self.width.get());
        let bin_index = (x / width).floor();

        // HOW TO HANDLE possible truncation from f64 to i64
        // only problematic for very large x and very small bin widths?
        let bin_index = bin_index as i64;

        self.data
            .entry(bin_index)
            .or_insert_with(|| MeanSD::new(StandardDeviationMode::Sample))
            .update(x);
    }
}

/// Getters.
impl Binned {
    /// Returns the bin width.
    #[must_use]
    pub const fn width(&self) -> NonZeroU32 {
        self.width
    }

    /// Returns a reference to the stats for all observations.
    #[must_use]
    pub const fn all(&self) -> &MeanSD {
        &self.all
    }

    /// Returns an iterator over the bins.
    pub fn bins(&self) -> impl Iterator<Item = (Bin, &MeanSD)> {
        self.data.iter().map(move |(index, meansd)| {
            let bottom = i64::from(self.width.get()) * index;
            let top = i64::from(self.width.get()) * (index + 1) - 1;

            (Bin { bottom, top }, meansd)
        })
    }
}

// ----------------------------------------------------------------------------
// tests
// ----------------------------------------------------------------------------

// silences Default::default warning from approx_eq macro
// may be fixed in https://github.com/mikedilger/float-cmp/pull/26
#[allow(clippy::default_trait_access)]
#[cfg(test)]
mod tests {
    use super::*;

    use float_cmp::approx_eq;

    #[test]
    fn positive() {
        let width = NonZeroU32::new(5).unwrap();
        let mut binned = Binned::new(width, StandardDeviationMode::Sample);

        binned.update(2.0);
        binned.update(3.0);
        binned.update(4.0);
        binned.update(5.0);
        binned.update(6.0);
        binned.update(7.0);

        {
            let (n, mean, sd) = binned.data.get(&0).unwrap().tupled();
            assert_eq!(n, 3);
            assert!(approx_eq!(f64, 3.0, mean));
            assert!(approx_eq!(f64, 1.0, sd));
        }

        {
            let (n, mean, sd) = binned.data.get(&1).unwrap().tupled();
            assert_eq!(n, 3);
            assert!(approx_eq!(f64, 6.0, mean));
            assert!(approx_eq!(f64, 1.0, sd));
        }

        {
            let (n, mean, sd) = binned.all.tupled();
            assert_eq!(n, 6);
            assert!(approx_eq!(f64, 4.5, mean));
            assert!(approx_eq!(f64, 1.870_828_693_386_970_7, sd));
        }
    }

    #[test]
    fn negative() {
        let width = NonZeroU32::new(5).unwrap();
        let mut binned = Binned::new(width, StandardDeviationMode::Sample);

        binned.update(-3.0);
        binned.update(-4.0);
        binned.update(-5.0);
        binned.update(-6.0);
        binned.update(-7.0);
        binned.update(-8.0);

        {
            let (n, mean, sd) = binned.data.get(&-1).unwrap().tupled();
            assert_eq!(n, 3);
            assert!(approx_eq!(f64, -4.0, mean));
            assert!(approx_eq!(f64, 1.0, sd));
        }

        {
            let (n, mean, sd) = binned.data.get(&-2).unwrap().tupled();
            assert_eq!(n, 3);
            assert!(approx_eq!(f64, -7.0, mean));
            assert!(approx_eq!(f64, 1.0, sd));
        }

        {
            let (n, mean, sd) = binned.all.tupled();
            assert_eq!(n, 6);
            assert!(approx_eq!(f64, -5.5, mean));
            assert!(approx_eq!(f64, 1.870_828_693_386_970_7, sd));
        }
    }
}
