use std::fmt;

use crate::StandardDeviationMode;

/// An online algorithm for calculating mean and standard deviation.
///
/// The memory requirement is constant.
#[derive(Clone, Copy, Debug, Default)]
pub struct MeanSD {
    sd_mode: StandardDeviationMode,
    size: u64,
    mean: f64,
    m2: f64,
}

/// Constructors and getters.
impl MeanSD {
    /// Returns an empty `MeanSD` structure.
    #[must_use]
    pub fn new(sd_mode: StandardDeviationMode) -> Self {
        Self {
            sd_mode,
            ..Self::default()
        }
    }
}

/// Add new observations.
impl MeanSD {
    /// Updates the data structure with a new observation `x`.
    pub fn update(&mut self, x: f64) {
        let delta = x - self.mean;

        self.size += 1;
        self.mean += delta / self.size_f();
        self.m2 += delta * (x - self.mean);
    }
}

/// Statistics related functions.
impl MeanSD {
    /// Returns the number of observations.
    #[must_use]
    pub const fn size(&self) -> u64 {
        self.size
    }

    /// Returns the mean.
    #[must_use]
    pub const fn mean(&self) -> f64 {
        self.mean
    }

    /// Returns the standard deviation.
    #[must_use]
    pub fn sd(&self) -> f64 {
        if self.size > 1 {
            let div = match self.sd_mode {
                StandardDeviationMode::Population => self.size_f(),
                StandardDeviationMode::Sample => self.size_f() - 1.0,
            };

            (self.m2 / div).sqrt()
        } else {
            0.0
        }
    }

    /// Returns the standard error of the mean.
    #[must_use]
    pub fn sem(&self) -> f64 {
        self.sd() / self.size_f().sqrt()
    }
}

/// Internals.
impl MeanSD {
    /// Returns size as [`f64`] for computation.
    const fn size_f(&self) -> f64 {
        // HOW TO HANDLE loss of precision from u64 to f64
        // only problematic for very large number of observations?
        self.size as f64
    }

    #[cfg(test)]
    pub(crate) fn tupled(&self) -> (u64, f64, f64) {
        (self.size, self.mean, self.sd())
    }
}

/// Formats to `"n={size} ∅ {mean} ± {sd}"`.
impl fmt::Display for MeanSD {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let n = self.size();
        let mean = self.mean().round();
        let sd = self.sd().round();

        write!(f, "n={} \u{2205} {} \u{b1} {}", n, mean, sd)
    }
}

// ----------------------------------------------------------------------------
// tests
// ----------------------------------------------------------------------------

// silences Default::default warning from approx_eq macro
// may be fixed in https://github.com/mikedilger/float-cmp/pull/26
#[allow(clippy::default_trait_access)]
#[cfg(test)]
mod tests {
    use super::*;

    use float_cmp::approx_eq;

    #[test]
    fn empty() {
        let meansd = MeanSD::default();
        let (n, mean, sd) = meansd.tupled();

        assert_eq!(n, 0);
        assert!(approx_eq!(f64, 0.0, mean));
        assert!(approx_eq!(f64, 0.0, sd));
    }

    #[test]
    fn single() {
        let mut meansd = MeanSD::default();

        meansd.update(1.0);

        let (n, mean, sd) = meansd.tupled();

        assert_eq!(n, 1);
        assert!(approx_eq!(f64, 1.0, mean));
        assert!(approx_eq!(f64, 0.0, sd));
    }

    #[test]
    fn small_positive() {
        let mut meansd = MeanSD::default();

        meansd.update(1.0);
        meansd.update(2.0);
        meansd.update(3.0);

        let (n, mean, sd) = meansd.tupled();

        assert_eq!(n, 3);
        assert!(approx_eq!(f64, 2.0, mean));
        assert!(approx_eq!(f64, 1.0, sd));
    }

    #[test]
    fn small_negative() {
        let mut meansd = MeanSD::default();

        meansd.update(-1.0);
        meansd.update(-2.0);
        meansd.update(-3.0);

        let (n, mean, sd) = meansd.tupled();

        assert_eq!(n, 3);
        assert!(approx_eq!(f64, -2.0, mean));
        assert!(approx_eq!(f64, 1.0, sd));
    }

    #[test]
    fn small_mixed() {
        let mut meansd = MeanSD::default();

        meansd.update(-1.0);
        meansd.update(0.0);
        meansd.update(1.0);

        let (n, mean, sd) = meansd.tupled();

        assert_eq!(n, 3);
        assert!(approx_eq!(f64, 0.0, mean));
        assert!(approx_eq!(f64, 1.0, sd));
    }

    #[test]
    fn population() {
        let mut meansd = MeanSD::new(StandardDeviationMode::Population);

        meansd.update(2.0);
        meansd.update(4.0);
        meansd.update(4.0);
        meansd.update(4.0);
        meansd.update(5.0);
        meansd.update(5.0);
        meansd.update(7.0);
        meansd.update(9.0);

        let (n, mean, sd) = meansd.tupled();

        assert_eq!(n, 8);
        assert!(approx_eq!(f64, 5.0, mean));
        assert!(approx_eq!(f64, 2.0, sd));
    }
}
